#!/usr/bin/env python

### Cesar Rodriguez Aug 2021  
### Trajectory controller     
### Based off of Klausen 2017 
#~~~ Uses generated trajectory and feedback to minimize oscillations
#~~~ Controller determines desired forces on drone to dampen oscillations 
#~~~ which are converted into attitude commands

import rospy,  tf
import numpy as np
import time
import math
import rosservice

from tf.transformations   import *
from scipy.integrate  	  import odeint 
from oscillation_ctrl.msg import RefSignal, LoadAngles
from oscillation_ctrl.srv import WaypointTrack
from geometry_msgs.msg 	  import Pose, Point, TwistStamped, PoseStamped
from sensor_msgs.msg 	  import Imu
from mavros_msgs.msg 	  import AttitudeTarget
from collections          import deque

class DesiredPoint(Point):
	def __init__(self,x,y,z):
		self.x = x
		self.y = y
		self.z = z

class LiveFilter:
    """Base class for live filters.
    """
    def process(self, x):
        # do not process NaNs
        # if numpy.isnan(x):
        #     return x

        return self._process(x)

    def __call__(self, x):
        return self.process(x)

    def _process(self, x):
        raise NotImplementedError("Derived class must implement _process")

class LiveLFilter(LiveFilter):
    def __init__(self, b, a):
        """Initialize live filter based on difference equation.

        Args:
            b (array-like): numerator coefficients obtained from scipy.
            a (array-like): denominator coefficients obtained from scipy.
            n (int): How many dimensions are we dealing with?
        """
        self.b = b
        self.a = a
        n = len(b)
        # 3 because x,y,z
        # self.n because that is how many terms we need
        self._xs = deque(np.zeros(n), maxlen = n) # unfiltered data
        self._ys = deque(np.zeros(n - 1), maxlen = n - 1) # filtered signal

    def _process(self, x):
        """
        Filter incoming data with standard difference equations*:
            a_0*y[n] = b_0*x[n] + b_1*x[n-1] + b_2*x[n-2] - a_1*y[n-1] - a_2*y[n-2]
            *Changes depending on order of filter
        """
        # y = numpy.empty(3)
        # self._xs.appendleft(x)
        # # get every x val (idx = 0), then y val (idx=1), then z val (idx=2) to determine y
        # for i in range(3):
        #     last_x_vals = [val_x[i] for val_x in self._xs]
        #     last_y_vals = [val_y[i] for val_y in self._ys]
        #     y[i] = numpy.dot(self.b, last_x_vals) - numpy.dot(self.a[1:], last_y_vals)

        # y = y / self.a[0]
        # self._ys.appendleft(y)
        
        # return y

        self._xs.appendleft(x)
        y = np.dot(self.b, self._xs) - np.dot(self.a[1:], self._ys)
        y = y / self.a[0]
        self._ys.appendleft(y)

        return y

class AltCtrlr():

	def __init__(self, Kp, Ki, Kd, drone_mass):
		self.Kp = Kp
		self.Ki = Ki
		self.Kd = Kd
		self.m = drone_mass

		self.e = 0.0
		self.e_p = 0.0
		self.edot = 0.0
		self.esum = 0.0

		self.t_prev = 0.0

		self.use_ki = False

	# def update_thrust(self, pos_des, pos, dt):
	def update_thrust(self, pos_des, pos, orientation, t_curr):
		""" determines new thrust based on error"""

		try:
			# determine error
			self.e = pos_des - pos
			self.edot = (self.e - self.e_p) / (t_curr - self.t_prev)
			# only accumulate error if we are actually using ths thrust controller
			if self.use_ki:
				self.esum += self.e * (t_curr - self.t_prev)
			else:
				self.esum = 0.0
			# get thrust
			# this formula ignores the path acc
			thrust = (self.Kp * self.e + self.Ki * self.esum + self.Kd * self.edot + self.m * 9.81) / (math.cos(orientation[0]) * math.cos(orientation[1])) 
			
			# update values
			self.e_p = self.e
			self.t_prev = t_curr
			
			return thrust
		except ZeroDivisionError:
			print("ZeroDivisionError in update_thrust")
			return 0.0

class Main:

	def __init__(self):

		# rate(s)
		rate = 20 # rate for the publisher method, specified in Hz -- 50 Hz #25

		# initialize variables
		
		# time variables
		self.dt 	= 1.0/rate
		self.tmax	= self.dt
		self.n		= self.tmax/self.dt + 9
		self.t 		= np.linspace(0, self.tmax, self.n) # Time array
		self.tstart	= rospy.get_time() # Keep track of the start time
		while self.tstart == 0.0:				 # Need to make sure get_rostime works
			self.tstart = rospy.get_time()
		
		# Msgs types
		self.vel_data	 = TwistStamped() # This is needed to get drone vel from gps
		self.imu_data	 = Imu() 				 # Needed for to get drone acc from IMU
		self.path_pos	 = np.zeros([3,1])
		self.path_vel	 = np.zeros([3,1]) 
		self.path_acc 	 = np.zeros([3,1]) 
		self.dr_pos 	 = Pose()
		self.att_targ    = AttitudeTarget()	# used to send quaternion attitude commands
		self.load_angles = LoadAngles()
		self.EulerAng	 = [0,0,0] # Will find the euler angles, and then convert to q
		self.EulerPose   = [0,0,0]

		# Service var
		self.get_xd = rospy.ServiceProxy('/status/waypoint_tracker',WaypointTrack)		
		self.empty_point = Point() # Needed to query waypoint_server

		# Drone var
		self.has_run	= False # Bool to keep track of first run instance

		# Col1 = theta, theta dot; Col2 = phi, phidot for self.PHI
		self.PHI 		= np.array([[0,0],[0,0]]) 
		self.dr_vel		= np.zeros([3,1])
		self.dr_angVel	= np.zeros([3,1])
		self.dr_acc		= np.zeros([3,1]) #self.imu_data.linear_acceleration

		# Tether var - Check if current method is used
		# Get tether length
		if rospy.get_param('status/pload', False):
			self.tetherL = self.get_tether()
			if self.tetherL == 0.0:
				self.tether = False
			else:
				self.tether = True
		else: 
			self.tetherL = 0.0
			self.tether = False

		# Retrieve drone and payload masses from config file
		[self.drone_m, self.pl_m] = self.get_masses()
		rospy.loginfo('DRONE MASS: %.2f',self.drone_m)
		self.tot_m 	  = self.drone_m + self.pl_m

		# Values which require updates. *_p = prior
		self.z1_p	  =	np.zeros([3,1])
		self.alpha	  = np.zeros([5,1])
		self.alphadot = np.zeros([5,1])
		self.a45	  = np.zeros(2)
		self.a45dot	  = np.array([[0],[0]])

		# Error terms
		self.z1		= np.zeros([3,1]) # dr_pos - ref_sig_pos
		self.z2 	= np.zeros([5,1]) # [vx;vy;vz;thetadot;phidot] - alpha
		self.z_sum  = np.zeros([3,1])
	
		# Tuning gains
		self.K1	= np.identity(3)*1#*0.5 # 2.0
		self.K2	= np.identity(5)*1#2.15#1.5#*0.1 # increases velocity and acceleration, also increases pos error

		# self.tune = 0.005 # Tuning parameter
		self.tune = 0.5 # Tuning parameter 0.1 gives better performance, 0.2 for Gazebo
		self.K2tune = 10 #1000 # 2023-04-18: Works great in Gazebo. Needs to be x1000 greater than the first 3 terms of K2
		# self.K2tune = 0.1 # Tuning parameter for last 3 terms of K2
		self.dist =	np.array([0,0,0,0.1,0.1]) # Wind disturbance 
		# self.dist =	np.array([0,0,0,0.0,0.0]) # Wind disturbance 
		# Gain terms
		self.Kp = np.identity(3) + np.dot(self.K2[:3,:3],self.K1) + self.tune*np.identity(3)
		self.Kd = self.tot_m*self.K1 + self.K2[:3,:3] #+ self.tune*self.K2[:3,:3]
		self.Ki = self.tune*self.K1

		# PD Thrust Controller terms
		# gains for thrust PD Controller
		### Failed when both gains were set to 1.0
		# self.Kp_thrust 	= 5 # 1.5 for lab # 3.5 for gazebo 
		# self.Kp_thrust 	= 3.0 # 3.5 for PD controller not using trajectory tracking
		# self.Kd_thrust 	= 0.14 # 1.5 for lab # 2.5 for gazebo 2.0 if no tether?
		# self.Kd_thrust 	= 3.5 # 1.5 for lab # 2.5 for gazebo 2.0 if no tether?

		self.alt_ctrl = AltCtrlr(Kp=3.0, Kd=1.75, Ki=2.0, drone_mass=self.tot_m)

		self.R 	= np.empty([3,3]) # rotation matrix
		self.g = np.array([0,0,9.81]).reshape(3,1)
		self.e3 = np.array([[0],[0],[1]])
		self.error_p = 0.0
		self.error_vel = 0.0
		self.t_curr_t = rospy.get_time() + 0.1
		self.t_prev_t = rospy.get_time()
		# Get scaling thrust factor, kf
		self.kf = self.get_kf()
		self.thrust = 0.0

		self.service_list = rosservice.get_service_list()
# --------------------------------------------------------------------------------#		
# 									SUBSCRIBERS
# --------------------------------------------------------------------------------#
		# Topic, msg type, and class callback method
		rospy.Subscriber('/status/load_angles', LoadAngles, self.loadAngles_cb)
		rospy.Subscriber('/reference/path', RefSignal, self.refsig_cb)
		rospy.Subscriber('/mavros/local_position/pose', PoseStamped, self.dronePos_cb)
		rospy.Subscriber('/mavros/local_position/velocity_body', TwistStamped, self.droneVel_cb)
		rospy.Subscriber('/mavros/imu/data', Imu, self.droneAcc_cb)
# --------------------------------------------------------------------------------#		
# 									PUBLISHERS
# --------------------------------------------------------------------------------#
		self.pub_att_targ = rospy.Publisher('/command/att_target', AttitudeTarget, queue_size=10)

		# timer(s), used to control method loop freq(s) as defined by the rate(s)
		# rospy.Timer(rospy.Duration(2.0/rate), self.publisher) # this was 5.0 rate before
		# self.pub_time  = rospy.Timer(rospy.Duration(1.0/rate), self.publisher) # this was 5.0 rate before
		rospy.Timer(rospy.Duration(0.1/rate), self.publisher) # this was 5.0 rate before

		rospy.Timer(rospy.Duration(2.0/rate), self.determine_throttle_new) # this was 5.0 rate before
		# self.throttle_timer  = rospy.Timer(rospy.Duration(1.0/rate), self.determine_throttle) # this was 5.0 rate before
		# rospy.Timer(rospy.Duration(0.1/rate), self.determine_throttle) # this was 5.0 rate before # best performance when this runs faster than attitude ctrl

		# timer for user feedback
		rospy.Timer(rospy.Duration(0.1), self.user_feedback)

		# timer to check if we are using attitude controller
		self.timer = rospy.Timer(rospy.Duration(0.25), self.ctrl_cb)
		self.att_ctrl = False

	def ctrl_cb(self, timer):
		""" Check if we are using attitude controller.
			This is done to stop querying ROS parameters in each loop
		"""
		try:
			if rospy.get_param('status/use_ctrl'):
				self.att_ctrl = True
				self.timer.shutdown() # shut down timer once we it is true
		except KeyError:
			pass
			# rospy.loginfo('Waiting for params to be ready')
# --------------------------------------------------------------------------------#		
# 								SETUP PARAM METHODS
# --------------------------------------------------------------------------------#

	def get_tether(self):
		""" Get tether length based on set parameters"""
		param_exists = False
		while param_exists == False:
			rospy.loginfo('Waiting for tether length...')
			if rospy.has_param('status/tether_length'):
				tether_length	= rospy.get_param('status/tether_length') # Tether length
				rospy.loginfo('TETHER LENGTH IN CONFG FILE')
				param_exists = True
			elif rospy.get_time() - self.tstart >= 10.0:
				tether_length	= 0.0
				rospy.loginfo('TETHER LENGTH NOT FOUND IN PARAMS')
				break
		return tether_length

	def get_masses(self):
		""" Use parameters to retrieve drone and payload mass set in config file and sets total mass of bodies"""
		# Constants
		param_exists = False # reset to False to get body masses
		while param_exists == False:
			if rospy.has_param('status/drone_mass'):
				drone_m	= rospy.get_param('status/drone_mass') # wait time
				param_exists = True
			elif rospy.get_time() - self.tstart >= 5.0:
				drone_m = 0.5841
				rospy.loginfo('DRONE MASS NOT FOUND IN CONFIG FILE')
				break
		if self.tether:
			param_exists = False
			while param_exists == False:
				rospy.loginfo('Waiting for pload mass...')
				if rospy.has_param('status/pload_mass'):
					pl_m	= rospy.get_param('status/pload_mass') # wait time
					rospy.loginfo('PLOAD MASS FOUND')
					param_exists = True
				elif rospy.get_time() - self.tstart >= 5.0:
					pl_m = 0.0
					rospy.loginfo('PLOAD MASS NOT FOUND IN CONFIG FILE')
					break
		else:
			pl_m = 0.0

		rospy.loginfo('PLOAD MASS: %.2f',pl_m)
		return [drone_m, pl_m]
		
	def get_kf(self):
		if rospy.has_param('status/hover_throttle'):

			rospy.loginfo('total mass: %.3f',self.tot_m)

			hover_throttle = rospy.get_param('status/hover_throttle')
			rospy.loginfo('Using hover throttle from config %.3f',hover_throttle)
			self.max_throttle = 0.45
		else:

			rospy.loginfo('total mass: %.3f',self.tot_m)

			hover_throttle = (self.tot_m*9.81 + 8.9)/34.84 # linear scaling dependent on mass

			rospy.set_param('status/hover_throttle',hover_throttle)
			rospy.loginfo('Determined hover throttle to be %.3f',hover_throttle)
			self.max_throttle = 0.8
		kf = hover_throttle/(self.tot_m*9.81)
		rospy.set_param('status/motor_constant',kf)
		return kf

# --------------------------------------------------------------------------------#		
# 								CALLBACK FUNCTIONS
# --------------------------------------------------------------------------------#

	# Callback pload deflection angles and vel
	def loadAngles_cb(self,msg):
		try:
			self.load_angles = msg
			# Populate self.PHI
			# self.PHI = np.array([[self.load_angles.theta,self.load_angles.phi],[self.load_angles.thetadot,self.load_angles.phidot]])
			self.PHI = np.array([[self.load_angles.phi,self.load_angles.theta],[self.load_angles.phidot,self.load_angles.thetadot]])

			# self.PHI = np.array([[0.0,0.0],[0.0,0.0]]) # in case we want to test performance without load angles

			# print('PHI:\n{0}'.format(self.PHI))
			# print('Angles:\n{0}'.format(self.PHI[0,:]))
			# print('Ang_dot:\n{0}'.format(self.PHI[1,:]))

		except ValueError as e:
			rospy.loginfo('Load angles callback failed due to: {0}'.format(e))

	# Callback drone pose
	def dronePos_cb(self,msg): 
		""" Subscribed to /mavros/local_position/pose, gets PoseStamped msgs """
		try:
			self.dr_pos = msg.pose
			self.EulerPose = self.convert2eul(self.dr_pos.orientation)
			# self.dr_pos = msg.drone_pos
		except ValueError as e:
			rospy.loginfo('Drone pos callback failed due to: {0}'.format(e))

	# Callback for drone velocity
	def droneVel_cb(self,msg):
		try:
			self.dr_vel 	= np.array([msg.twist.linear.x,msg.twist.linear.y,msg.twist.linear.z]).reshape(3,1)
			self.dr_angVel	= np.array([[msg.twist.angular.x],[msg.twist.angular.y],[msg.twist.angular.z]])

		except (ValueError or TypeError) as e:
			rospy.loginfo('Drone vel callback failed due to: {0}'.format(e))

	# Callback for drone accel from IMU data
	def droneAcc_cb(self,msg):
		try:
			self.dr_acc = np.array([[msg.linear_acceleration.x],[msg.linear_acceleration.y],[msg.linear_acceleration.z]])

		except ValueError as e:
			rospy.loginfo('Drone accel. callback failed due to: {0}'.format(e))

	# Callback reference signal
	def refsig_cb(self,msg):
		try:
			self.path_pos = np.array([[msg.position.x],[msg.position.y],[msg.position.z]])
			self.path_vel = np.array([[msg.velocity.x],[msg.velocity.y],[msg.velocity.z]])
			self.path_acc =	np.array([[msg.acceleration.x],[msg.acceleration.y],[msg.acceleration.z]])

		except ValueError as e:
			rospy.loginfo('Reference signal callback failed due to: {0}'.format(e))

	def waypoints_srv_cb(self):
		if '/status/waypoint_tracker' in self.service_list:
			rospy.wait_for_service('/status/waypoint_tracker')
			try:
				resp = self.get_xd(False,self.empty_point)
				self.xd = resp.xd
			except ValueError as e:
				rospy.loginfo('{0}'.format(e))
		else:
			self.xd = Point(x=0.0,y=0.0,z=2.0)

# ---------------------------------ODE SOLVER-------------------------------------#
	def statespace(self,y,t,Ka,Kb,Kc):
		# Need the statespace array:
		_y = np.array(y).reshape(2,1)
		# print('Ka:\n{0}\nKb:\n{1}'.format(Ka, Kb))		

		K = np.dot(Ka,_y) + Kb # Kb should be 2x1

		# Derivative of statespace array:
		dydt = np.dot(Kc,K)
		# print('dydt:\n{0}'.format(dydt))		
		dydt = [dydt[0][0], dydt[1][0]]
		
		return dydt
# --------------------------------------------------------------------------------#
	def quaternion_rotation_matrix(self):
		"""
		Covert a quaternion into a full three-dimensional rotation matrix.

		Input
		:param Q: A 4 element array representing the quaternion (q0,q1,q2,q3) 

		Output
		:return: A 3x3 element matrix representing the full 3D rotation matrix. 
		         This rotation matrix converts a point in the local reference 
		         frame to a point in the global reference frame.
		"""
		# Extract the values from Q
		q0 = self.dr_pos.orientation.w
		q1 = self.dr_pos.orientation.x
		q2 = self.dr_pos.orientation.y
		q3 = self.dr_pos.orientation.z
		 
		# First row of the rotation matrix
		r00 = 2 * (q0 * q0 + q1 * q1) - 1
		r01 = 2 * (q1 * q2 - q0 * q3)
		r02 = 2 * (q1 * q3 + q0 * q2)
		 
		# Second row of the rotation matrix
		r10 = 2 * (q1 * q2 + q0 * q3)
		r11 = 2 * (q0 * q0 + q2 * q2) - 1
		r12 = 2 * (q2 * q3 - q0 * q1)
		 
		# Third row of the rotation matrix
		r20 = 2 * (q1 * q3 - q0 * q2)
		r21 = 2 * (q2 * q3 + q0 * q1)
		r22 = 2 * (q0 * q0 + q3 * q3) - 1
		 
		# 3x3 rotation matrix
		rot_matrix = np.array([[r00, r01, r02],
		                       [r10, r11, r12],
		                       [r20, r21, r22]])
		                        
		return rot_matrix

	def convert2eul(self,quaternion_orientation):
		"""
		Convers quaternion in pose message into euler angles

		Input
		:param Q: orientation pose message

		Output
		:return:  Array of euler angles
		"""
		x = quaternion_orientation.x
		y = quaternion_orientation.y
		z = quaternion_orientation.z
		w = quaternion_orientation.w
		
		t0 = +2.0 * (w * x + y * z)
		t1 = +1.0 - 2.0 * (x * x + y * y)
		roll = math.atan2(t0, t1)

		t2 = +2.0 * (w * y - z * x)
		t2 = +1.0 if t2 > +1.0 else t2
		t2 = -1.0 if t2 < -1.0 else t2
		pitch = math.asin(t2)

		t3 = +2.0 * (w * z + x * y)
		t4 = +1.0 - 2.0 * (y * y + z * z)
		yaw = math.atan2(t3, t4)
		
		euler = [roll, pitch, yaw]
		return euler

	def determine_throttle_old(self,throttle_timer):
		""" Thrust as per Geometric Tracking Control of a Quadrotor UAV on SE(3) Taeyoung Lee, Melvin Leok, and N. Harris McClamroch"""

		self.waypoints_srv_cb() # check for new waypoints
		self.R = self.quaternion_rotation_matrix() # get rotation matrix
		b3 = self.R.dot(np.array([0,0,1]).reshape(3,1))

		self.error = np.array([[self.path_pos[0] - self.dr_pos.position.x],
							   [self.path_pos[1] - self.dr_pos.position.y],
							   [self.path_pos[2] - self.dr_pos.position.z]]).reshape(3,1)

		self.error_vel = self.path_vel.reshape(3,1) - self.dr_vel # best performance

		thrust_vector = self.tot_m*(self.g + self.path_acc) + self.Kp_thrust*self.error + self.Kd_thrust*self.error_vel
		# thrust_vector = (self.g*self.tot_m + self.Kp_thrust*self.error + self.Kd_thrust*self.error_vel) * self.kf

		self.thrust = thrust_vector/(math.cos(self.EulerPose[0])*math.cos(self.EulerPose[1]))
		# self.thrust = (thrust_vector[0] * b3[0] + thrust_vector[1] * b3[1] + thrust_vector[2] * b3[2]) * self.kf # best performance
		# self.thrust = (thrust_vector[0] * b3[0] + thrust_vector[1] * b3[1] + thrust_vector[2] * b3[2]) # best performance
		# thrust = (thrust_vector[0] * b3[0] + thrust_vector[1] * b3[1] + thrust_vector[2] * b3[2]) * self.kf # best performance

		# Value needs to be between 0 - 1.0
		self.att_targ.thrust = max(0.0,min(self.thrust * self.kf,self.max_throttle)) # will never want it to be more than 0.5 with LabDrone
		# self.att_targ.thrust = max(0.0,min(thrust,self.max_throttle)) # will never want it to be more than 0.5 with LabDrone
		# print('thrust_vector:\n{0}\nthrust:{1}'.format(thrust_vector, self.att_targ.thrust))

		# Fd_tf = quaternion_multiply(dr_orientation,quaternion_multiply([Fd[0],Fd[1],self.thrust,0.0],dr_orientation_inv)) # Real part of Fd needs = 0

	def determine_throttle(self,throttle_timer):
		""" Thrust as per Geometric Tracking Control of a Quadrotor UAV on SE(3) Taeyoung Lee, Melvin Leok, and N. Harris McClamroch"""

		self.waypoints_srv_cb() # check for new waypoints
		# self.R = self.quaternion_rotation_matrix() # get rotation matrix
		# b3 = self.R.dot(np.array([0,0,1]).reshape(3,1))
		# b3 = np.array([0,0,1])
		
		self.t_curr_t = rospy.get_time()

		self.error = np.array([[0],
							   [0],
							   [self.path_pos[2] - self.dr_pos.position.z]]).reshape(3,1)
		self.e_sum += self.error

		# self.error_vel = self.path_vel.reshape(3,1) - self.dr_vel # best performance
		self.error_vel = (self.error[2] - self.error_p) / (self.t_curr_t - self.t_prev_t) 


		# thrust_vector = self.tot_m*(self.g[2] + self.path_acc[2]) + self.Kp_thrust*self.error[2] + self.Kd_thrust*self.error_vel[2]
		thrust_vector = self.tot_m*(self.g[2] + self.path_acc[2]) + self.Kp_thrust*self.error[2] + self.Kd_thrust*self.error_vel

		# thrust_vector = self.tot_m*(self.g + self.path_acc) + self.Kp_thrust*self.error + self.Kd_thrust*self.error_vel
		# thrust_vector = (self.g*self.tot_m + self.Kp_thrust*self.error + self.Kd_thrust*self.error_vel) * self.kf

		# thrust = thrust_vector/(math.cos(self.EulerPose[0])*math.cos(self.EulerPose[1]))
		self.thrust = thrust_vector/(math.cos(self.EulerPose[0])*math.cos(self.EulerPose[1]))
		# self.thrust = (thrust_vector[0] * b3[0] + thrust_vector[1] * b3[1] + thrust_vector[2] * b3[2]) * self.kf # best performance
		# self.thrust = (thrust_vector[0] * b3[0] + thrust_vector[1] * b3[1] + thrust_vector[2] * b3[2]) # best performance
		# thrust = (thrust_vector[0] * b3[0] + thrust_vector[1] * b3[1] + thrust_vector[2] * b3[2]) * self.kf # best performance

		# Value needs to be between 0 - 1.0
		self.att_targ.thrust = max(0.0, min(self.thrust * self.kf, self.max_throttle)) # will never want it to be more than 0.5 with LabDrone
		# self.att_targ.thrust = max(0.0,min(thrust,self.max_throttle)) # will never want it to be more than 0.5 with LabDrone
		# print('thrust_vector:\n{0}\nthrust:{1}'.format(thrust_vector, self.att_targ.thrust))

		# update
		self.error_p = self.error[2]
		self.t_prev_t = self.t_curr_t

	def determine_throttle_new(self, throttle_timer):
		self.waypoints_srv_cb() # check for new waypoints

		# check if we are using the controller
		self.alt_ctrl.use_ki = self.att_ctrl

		# self.t_curr_t = rospy.get_time()
		self.thrust = self.alt_ctrl.update_thrust(self.path_pos[2], self.dr_pos.position.z, self.EulerPose ,rospy.get_time())
		# self.thrust = self.alt_ctrl.update_thrust(self.path_pos[2], self.dr_pos.position.z, rospy.get_time())/(math.cos(self.EulerPose[0])*math.cos(self.EulerPose[1]))
		self.att_targ.thrust = max(0.0, min(self.thrust * self.kf, self.max_throttle)) # will never want it to be more than 0.5 with LabDrone
		# self.att_targ.thrust = max(0.0, min(self.thrust, self.max_throttle)) # will never want it to be more than 0.5 with LabDrone

	def determine_q(self):
		""" Determine attitude commands based on feedback and feedforward methods"""
		# Populate position vector and gamma (g). g is state space vector: [px,py,pz,theta,phi]
		p =	np.array([[self.dr_pos.position.x],[self.dr_pos.position.y],[self.dr_pos.position.z]])
		g = np.array([[self.dr_pos.position.x],[self.dr_pos.position.y],[self.dr_pos.position.z],[self.load_angles.theta],[self.load_angles.phi]])
 
		# State some variables for shorthand
		L 		= self.tetherL
		c_theta = math.cos(self.load_angles.theta)
		c_phi 	= math.cos(self.load_angles.phi)
		s_theta = math.sin(self.load_angles.theta)
		s_phi	= math.sin(self.load_angles.phi)

		# Control matrices
		M = [[self.tot_m, 0, 0, 0, L*self.pl_m*c_theta],
			 [0, self.tot_m, 0, -L*self.pl_m*c_phi*c_theta, L*self.pl_m*s_phi*s_theta],
			 [0, 0, self.tot_m, -L*self.pl_m*c_theta*s_phi, -L*self.pl_m*c_phi*s_theta],
			 [0, -L*self.pl_m*c_phi*c_theta,-L*self.pl_m*c_theta*s_phi, (L**2)*self.pl_m*c_theta**2 + 0.001*s_theta**2, 0],
			 [L*self.pl_m*c_theta, L*self.pl_m*s_phi*s_theta, -L*self.pl_m*c_phi*s_theta, 0, L**2*self.pl_m]]

		C = [[0,0,0,0,-L*self.load_angles.thetadot*self.pl_m*s_theta],
			 [0,0,0,L*self.pl_m*(self.load_angles.phidot*c_theta*s_phi + self.load_angles.thetadot*c_phi*s_theta), L*self.pl_m*(self.load_angles.phidot*c_phi*s_theta  + self.load_angles.thetadot*c_theta*s_phi)],
			 [0,0,0,-L*self.pl_m*(self.load_angles.phidot*c_phi*c_theta - self.load_angles.thetadot*s_phi*s_theta),-L*self.pl_m*(self.load_angles.thetadot*c_phi*c_theta - self.load_angles.phidot*s_phi*s_theta)],
			 [0,0,0,-0.5*(L**2)*self.pl_m*self.load_angles.thetadot*math.sin(2*self.load_angles.theta) + 0.5*0.001*self.load_angles.thetadot*math.sin(2*self.load_angles.theta), -0.5*(L**2)*self.pl_m*self.load_angles.phidot*math.sin(2*self.load_angles.theta)],
			 [0,0,0,0.5*(L**2)*self.pl_m*self.load_angles.phidot*math.sin(2*self.load_angles.theta),0]]

		G = [[0],[0],[9.81*self.tot_m],[L*9.81*self.pl_m*c_theta*s_phi],[L*9.81*self.pl_m*c_phi*s_theta]]
		# G = [[0],[0],[9.81*self.tot_m],[L*9.81*self.pl_m*c_theta*s_phi],[L*9.81*self.pl_m*c_phi*s_theta]] ##########

		H = [[1,0,0,0,0],[0,1,0,0,0],[0,0,1,0,0]]

		D = np.diag(self.dist)	# Already in array format

		# Need to convert to useful format -> np.array
		M = np.array(M)
		C = np.array(C)
		G = np.array(G)
		H = np.array(H)

		# Shorthand for matrices which will be used
		M_a  = M[3:5,3:5] # M_alpha - rows 4 to 5 and columns 4 to 5 ... looks wrong its not
		C_a  = C[3:5,3:5]
		D_a  = D[3:5,3:5]
		G_a  = G[3:5]
		M_b  = M[3:5,:3] # M_4:5,1:3 - rows 4 to 5 and columns 1 to 3
		M_c	 = M[:3,3:5] # M_1:3,4:5 - rows 1 to 3 and columns 4 to 5
		C_c  = C[:3,3:5]
	
		# Determine alpha
		# if not self.tether:
		if self.tether:
			# Constants from Eq. 49
			# Ka = -(D_a + C_a + self.K2[3:5,3:5])
			# Kb = -G_a + np.dot(self.K2[3:5,3:5],self.PHI[1,:].reshape(2,1)) - np.dot(M_b,self.path_acc - np.dot(self.K1,self.dr_vel - self.path_vel)) 
			Ka = -(D_a + C_a + self.K2tune*self.K2[3:5,3:5])
			Kb = -G_a + np.dot(self.K2tune*self.K2[3:5,3:5],self.PHI[1,:].reshape(2,1)) - np.dot(M_b,self.path_acc - np.dot(self.K1, self.dr_vel - self.path_vel)) 
# 
			M_aI = np.linalg.inv(M_a) # Inverse of M_a
			# print('M_aI was found to be:\n{0}\n'.format(M_aI))

			# SOLVE ODE (get alpha)
			# Populate buffer 
			self.a45_buff  = odeint(self.statespace,self.a45,self.t,args=(Ka,Kb,M_aI))
			# self.a45_buff  = odeint(self.statespace,self.a45,self.t,args=(Ka,Kb,np.identity(2))) # spits out a 2,2 matrix
			# print('a45_0 was found to be:\n{0}\na45 was found to be:\n{1}\n'.format(self.a45_0,self.a45))
			# self.a45dot = np.array(self.statespace(self.a45,1,Ka,Kb,M_aI)) # Do not need 't' and that's why it is a 1
			
			# Update a45_0 to new a45. Need to transpose to column vector
			# self.a45_0 = self.a45_buff[0,:]
			self.a45 = self.a45_buff[-1,:]
			# self.a45 = M_aI.dot(self.a45_buff[-1,:])

			# print('a45_0 was found to be {0}'.format(self.a45_0))

			# Get alphadot_4:5
			self.a45dot = np.array(self.statespace(self.a45,1,Ka,Kb,M_aI)) # Do not need 't' and that's why it is a 1

		else: # if no tether, alpha_4:5 = 0 
			self.a45	= np.array([[0],[0]])
			self.a45dot = np.array([[0],[0]])

		# populate error terms
		self.z1 = p.reshape(3,1) - self.path_pos.reshape(3,1)
		# z1_dot	= self.dr_vel - self.path_vel
		z1_dot	= self.R.dot(-self.path_vel) + self.dr_vel
		self.z2 = g - self.alpha

		# Determine alpha
		self.alpha[:3] = self.path_vel - np.dot(self.K1,self.z1)
		self.alpha[3:5] = self.a45.reshape(2,1)

		# determine stability factor
		B = np.dot(C_c,self.a45.reshape(2,1)) + np.dot(M_c,self.a45dot.reshape(2,1))

		dr_orientation     = [self.dr_pos.orientation.x, self.dr_pos.orientation.y, self.dr_pos.orientation.z, self.dr_pos.orientation.w]
		dr_orientation_inv = quaternion_inverse(dr_orientation)
		
		if self.att_ctrl:
			z_int = self.z_sum + self.z1 * self.dt
		else:
			z_int = np.zeros(3).reshape(3,1)

		# path_acc = self.R.dot(self.path_acc)
		# path_acc = np.array([self.path_acc[0],self.path_acc[1],self.path_acc[2]+9.81]).reshape(3,1) # need to add gravity for controller to work
		path_acc = np.array([self.path_acc[0],self.path_acc[1],self.path_acc[2]]).reshape(3,1) # need to add gravity for controller to work

		# Desired body-oriented forces
		Fd = B + G[:3] + self.tot_m*path_acc - np.dot(self.Kd,z1_dot) - np.dot(self.Kp,self.z1) -  np.dot(self.Ki,z_int)
		# rospy.loginfo('\nFd:\n{0}'.format(Fd))
		
		# Update self.z1_p for integration		
		self.z1_p = self.z1
		self.z_sum = z_int

		# Covert Fd into drone frame, do not need Fz since we have a seperate altitude controller  
		# Fd_tf = quaternion_multiply(dr_orientation,quaternion_multiply([Fd[0],Fd[1],-self.thrust,0.0],dr_orientation_inv)) # Real part of Fd needs = 0
		# Fd_tf = quaternion_multiply(dr_orientation,quaternion_multiply([Fd[0],Fd[1],self.thrust,0.0],dr_orientation_inv)) # Real part of Fd needs = 0
		# Fd_tf = quaternion_multiply(dr_orientation,quaternion_multiply([Fd[0],Fd[1],Fd[2],0.0],dr_orientation_inv)) # Real part of Fd needs = 0
		Fd_tf = quaternion_multiply(dr_orientation,quaternion_multiply([Fd[0],Fd[1],0.0,0.0],dr_orientation_inv)) # Real part of Fd needs = 0, ignoring F_z since assuming acc_z = 0 

		# rospy.loginfo('\nFd_tf:\n{0}'.format(Fd_tf))
		
		# Convert forces to attitude *EulerAng[2] = yaw = 0
		self.EulerAng[1] = -math.atan(-Fd_tf[0]/(self.tot_m*9.81)) # Pitch
		self.EulerAng[0] = -math.atan(Fd_tf[1]*math.cos(self.EulerAng[1])/(self.tot_m*9.81)) # Roll	

		q = quaternion_from_euler(self.EulerAng[0],self.EulerAng[1],self.EulerAng[2])

		# self.user_feedback(Fd,Fd_tf)

		# Populate msg variable
		# Attitude control
		self.att_targ.header.stamp = rospy.Time.now()
		self.att_targ.header.frame_id = '/map'
		self.att_targ.type_mask = 1|2|4 # ignore body rate command
		self.att_targ.orientation.x = q[0] 
		self.att_targ.orientation.y = q[1]
		self.att_targ.orientation.z = q[2]
		self.att_targ.orientation.w = q[3]
		self.att_targ.body_rate.x = 0.0
		# self.att_targ.body_rate.x = self.thrust # for debuggin
		self.att_targ.body_rate.y = 0.0
		self.att_targ.body_rate.z = 0.0
		# self.att_targ.thrust = max(0.0,min(Fd_tf[2] * self.kf,self.max_throttle)) # will never want it to be more than 0.5 with LabDrone
 		
	def user_feedback(self, timer):
	# def user_feedback(self,F_noTransform, F_Transform):
		rospy.loginfo('\n----------------------')
		rospy.loginfo('thrust: %.6f' % self.thrust)
		rospy.loginfo('Attitude Mode: {0}'.format(self.att_ctrl))
		rospy.loginfo('roll: %.2f', self.EulerAng[0]*180/3.14)
		rospy.loginfo('pitch: %.2f', self.EulerAng[1]*180/3.14)
		# rospy.loginfo('Fd before transform: %.2f, %.2f, %.2f', F_noTransform[0], F_noTransform[1], F_noTransform[2])
		# rospy.loginfo('Fd after transform: %.2f, %.2f, %.2f', F_Transform[0],F_Transform[1],F_Transform[2])
		# rospy.loginfo('a45 was found to be:{0}'.format(self.a45))
		# rospy.loginfo('a45dot was found to be:{0}'.format(self.a45dot))

	def publisher(self,pub_time):
		# self.determine_throttle()
		self.determine_q()
		self.pub_att_targ.publish(self.att_targ)

if __name__=="__main__":

	# Initiate ROS node
	rospy.init_node('trajectoryCtrl',anonymous=True)
	try:
		Main() 			# create class object
		rospy.spin() 	# loop until shutdown signal

	except rospy.ROSInterruptException:
	        pass
